void HandleNoteOn(byte channel, byte note, byte velocity)
{
  if (channel == midiChannel)
  {
    midiVelocity = velocity;

    if (!midiMode)
    {
      midiMode = true; // switch to listening to notes from incoming MIDI
      envelopeProgress = 255;
      for (byte j = 0; j < 4; j++)
        voice[j] = 255;
      for (byte k = 0; k < 10; k++)
        rawArpList[k] = 255;
    }

    if (!arp)
    {
      if (!monoMode) // 4 voice paraphonic mode
      {
        voiceCounter = 0;
        for (byte i = 0; i < 4 ; i ++)
        {
          if (voice[i] != 255)
            voiceCounter++;
        }
        if (velocity == 0 && voiceCounter != 0)
        {
          for (byte i = 0; i < 4; i++)
          {
            if (voice[i] == note - 60)
            {
              voice[i] = 255;
              voiceCounter--;
            }
          }
          if (voiceCounter == 0)
          {
            envelopeProgress = 3; // we're in release phase
            releaseStartTime = millis();
          }
        }
        else
        {
          setVeloModulation(velocity);
          if (voiceCounter < 4)
          {
            byte j = 0;
            while (voice[j] != 255 && j < 4) // go to the next voice if this voice isn't free
              j++;
            if (j < 4)
            {
              voice[j] = note - 60;
              //showValue(13, 1, voice[j]);
              envelopeTrigger = true;
              if (retrigger) // should we retrigger the LFO?
                lfoIndex = 0;
              if (voiceCounter == 0)
              {
                //voiceCounter++;
                for (byte k = 0; k < 4; k++)
                {
                  if (k != j)
                    muteVoice[k] = true;
                }
              }
              voiceCounter++;
            }
          }
        }
      }
      else // we're in mono mode
      {
        if (velocity != 0)
        {
          for (byte i = 0; i < 4; i++)
            voice[i] = note - 60;
          if (envelopeProgress > 2) // we're in envelope decay phase (3), or envelope is at rest (255)
          {
            envelopeTrigger = true;
            if (retrigger) // should we retrigger the LFO?
              lfoIndex = 0;
          }
        }
      }
    }
    else //we're in arp mode
    {
      byte counter = 0;
      while (rawArpList[counter] != 255 && counter < 9)
        counter++;
      if (rawArpList[counter] == 255)
      {
        rawArpList[counter] = note - 60;
        sortArp();
      }
    }
  }
}

void HandleNoteOff(byte channel, byte note, byte velocity)
{
  if (channel == midiChannel)
  {
    if (!arp)
    {
      if (!monoMode) // we're in 4-voice paraphonic mode
      {
        voiceCounter = 0;
        for (byte i = 0; i < 4 ; i ++)
        {
          if (voice[i] != 255)
            voiceCounter++;
        }

        for (byte i = 0; i < 4; i++)
        {
          if (voice[i] == note - 60)
          {
            voice[i] = 255;
            voiceCounter--;
          }
        }
        if (voiceCounter == 0)
        {
          envelopeProgress = 3; // we're in release phase
          releaseStartTime = millis();
        }
        //showValue(13, 1, voiceCounter);
      }
      else // we are in mono mode
      {
        if (note - 60 == voice[0]) // in mono mode all 4 voices are the same, so we only have to compare against voice[0]
        {
          envelopeProgress = 3; // we're in release phase
          releaseStartTime = millis();
        }
      }
    }
    else //we're in arp mode
    {
      byte counter = 0;
      while (rawArpList[counter] != (note - 60) && counter < 9)
        counter++;
      if (rawArpList[counter] == (note - 60))
      {
        rawArpList[counter] = 255;
        sortArp();
      }
    }
  }
}

void HandleCC(byte channel, byte number, byte value)
{
  if (channel == midiChannel)
  {
    if (number == 2) // breath data received
    {
      if (!monoMode)
      {
        monoMode = true;
        lastFilterCutoff = filterCutoff;
      }
      filterCutoff = value << 1; // change range 0 - 127 to 0 - 255
    }
    midiCC[number] = value;
  }
}

void HandleClock(void) // what to do for 24ppq clock pulses
{
  if (syncIn)
  {
    lastClock = millis();
    receivingClock = true;
    static unsigned long lastClockInMicros = 0;
    if ((inPulseCounter % 12) == 0 && !seqJustStarted) // on each 8th except the first
    {
      pulseCounter = 0;
      eighthCounter++;
    }
    else
      seqJustStarted = false;

    clockInMicros = micros();
    if (inPulseCounter != 0)
    {
      pulseMicros = micros() - lastClockInMicros;
      tmpBpm[inPulseCounter % 24] = 60000000 / (pulseMicros * 24);
    }
    if (inPulseCounter > 4)
    {
      int totalBpm = 0;
      for (byte i = 0; i < 24; i++)
        totalBpm += tmpBpm[i];

      if (bpm != totalBpm / 24 + 1)
      {
        bpm = totalBpm / 24 + 1; // +1 here because for some reason the BPM reading is off by one
        setBpm();
        if (menu == 230)
          showValue(5, 1, bpm);
      }
    }
    inPulseCounter++;
    lastClockInMicros = clockInMicros;
  }
}

void HandleStart(void)
{
  if (syncIn)
  {
    seqRunning = true;
    seqStep = 0; // rewind the sequence
    inPulseCounter = 0;
    pulseCounter = 0;
    eighthCounter = 0;
    longStep = true;
    cueNextSeq();
    seqJustStarted = true;
  }
}

void HandleStop(void)
{
  if (syncIn)
  {
    seqRunning = false;
    seqSendMidiNoteOffs = true;
    seqStep = 0; // rewind the sequence
    pulseCounter = 0;
    eighthCounter = 0;
    longStep = true;

    noteRelease();
  }
}

void sendMidi()
{
  if (midiOut)
  {
    // ARPEGGIATOR
    if (arpSendMidiNoteOn)
    {
      if (arpMidiOut != 255)
        midiA.sendNoteOff(arpMidiOut, 127, midiChannel);
      arpMidiOut = sortedArpList[arpPosition] + (arpOctaveCounter * 12) + 60;
      midiA.sendNoteOn(arpMidiOut, 127, midiChannel);
      arpSendMidiNoteOn = false;
    }

    else if (arpSendMidiNoteOff)
    {
      midiA.sendNoteOff(arpMidiOut, 127, midiChannel);
      arpMidiOut = 255;
      arpSendMidiNoteOff = false;
    }

    // SEQUENCER
    else if (seqSendMidiNoteOns)
    {
      for (byte i = 0; i < 4; i++)
      {
        if (seqMidiOff[i] == 255 && seqMidiOn[i] != 255)
        {
          midiA.sendNoteOn(seqMidiOn[i] + 60, outVelocity, midiChannel);
          seqMidiOff[i] = seqMidiOn[i];
        }
        else if (seqMidiOn[i] != 255)
        {
          midiA.sendNoteOff(seqMidiOff[i] + 60, 127, midiChannel);
          midiA.sendNoteOn(seqMidiOn[i] + 60, outVelocity, midiChannel);
          seqMidiOff[i] = seqMidiOn[i];
        }
      }
      seqSendMidiNoteOns = false;
    }

    else if (seqSendMidiNoteOffs)
    {
      for (byte i = 0; i < 4; i++)
      {
        if (seqMidiOff[i] != 255)
        {
          midiA.sendNoteOff(seqMidiOff[i] + 60, 127, midiChannel);
          seqMidiOff[i] = 255;
        }
      }
      seqSendMidiNoteOffs = false;
    }
  }
}

void checkForClock()
{
  if (receivingClock)
  {
    if ((millis() - lastClock) > 300)
      receivingClock = false;
  }
}

void checkThru()
{
  if (midiThruType == 0)
    midiA.turnThruOff();
  else if (midiThruType == 1)
    midiA.turnThruOn(midi::Full);
  else if (midiThruType == 2)
    midiA.turnThruOn(midi::DifferentChannel); // in conjunction with MIDI_CHANNEL_OMNI, this only allows system messages though like clock, stop, start
}

void setSyncType()
{
  switch (midiSync)
  {
    case 0:
      syncIn = false;
      midiClockOut = 0;
      break;
    case 1:
      syncIn = false;
      midiClockOut = 1;
      break;
    case 2:
      syncIn = true;
      midiClockOut = 0;
      break;
  }
}
